/*
 * Altera Byteblaster driver
 *
 * Copyright (c) 2001 Clive Davies <cdavies@altera.com>
 * Copyright (c) 2001 Andrew Draper <adraper@altera.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* 14 Apr 2006
   Modified by Damien Bardon < damien83@free.fr >
   Remove devfs support ( it is not supporter by kernel 2.6 anymore )
*/


/*
 * ByteBlaster is a device which controls an IEEE1149 JTAG chain via
 * a parallel port.  Depending on which chips are in the JTAG chain
 * this can be used for board testing, device programming and debugging.
 *
 * Interface
 * ---------
 *
 * ioctl(BYTEBLASTER_IOCGETHW) returns the version of the driver.
 * This is currently 0
 *
 * ioctl(BYTEBLASTER_IOCGETVER) returns the version of the hardware attached
 * to the parallel port.
 *    0 means no recognised hardware.
 *    0x100 means v1 byteblaster, hw revision 0
 *    0x101 means v1 byteblaster, hw revision 1
 *
 * ioctl(BYTEBLASTER_IOCDIRECT) sets the values output on TCK, TMS and TDI
 * and returns the values sampled from RTCK and TDO.
 *   WRITE:
 *    b0 = TCK value
 *    b1 = TMS value
 *    b6 = TDI value
 *   READ:
 *    b4 = RTCK value
 *    b7 = TDO value
 *
 * write() takes a sequence of bytes and outputs one TCK for each
 * byte supplied.
 *    b0 = 0 if the input should be sampled, 1 otherwise
 *    b1 = TMS value
 *    b6 = TDI value
 *    b7 = xor TDI with previous TDO
 *    other bits should be 0
 *
 *  For each byte output with b0 = 0 one byte will be stored in the input FIFO
 *
 * read() reads bytes from the input FIFO.  Each byte has meaning as follows:
 *    b4 = value of RTCK on the rising edge of TCK
 *    b7 = value of TDO on the rising edge of TCK
 *    other bits are undefined
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/sched.h>
#include <asm/uaccess.h>

#include <linux/parport.h>
#include <linux/ioctl.h>
#include <linux/miscdevice.h>

#define BYTEBLASTER_IOCGETHW  _IOR(0xBB, 0, unsigned int *)
#define BYTEBLASTER_IOCGETVER _IOR(0xBB, 1, unsigned int *)
#define BYTEBLASTER_IOCDIRECT _IOWR(0xBB, 2, unsigned int *)

#define BB_VER_1  0x100
#define BB_REV_0  0
#define BB_REV_1  1

#define READ_FIFO_SIZE 8196 /* must be a power of 2 */

struct byteblaster_t {
	struct pardevice * dev;
	unsigned long      flags;
	struct semaphore   port_mutex;

	char             * read_fifo;
	unsigned int       read_fifo_in;
	unsigned int       read_fifo_out;
	char               last_in;
};


#define BYTEBLASTER_BUSY	 0x0004
#define BYTEBLASTER_BUSY_BIT_POS 2


#define BYTEBLASTER_MAX 4

static struct byteblaster_t byteblaster[BYTEBLASTER_MAX];

static int byteblaster_major = 0;

static unsigned short byteblaster_check_hardware(struct parport * port)
{
	unsigned int result;
	/*
	 * On the rev 0 PCB, DATA7 (BIT7) is hardwired to the PAPER_END
	 * signal (BIT5). On the rev 1 PCB, DATA5 (BIT5) is hardwired to
	 * the ACK signal (BIT6) and DATA7 is connected to PAPER_END
	 * through the 74244.
	 */
	/*
	 * Check for rev 0, leaving port in idle
	 * state (b6,b1 high, b0 low)
	 */
	/* Set data bits 5&7 low */
	parport_write_data (port, 0x42);
	/* Read bits 6&7 of the status register */
	result = parport_read_status (port);

	/* Set data bits 5&7 high */
	parport_write_data (port, 0xf2);
	/* Read bits 6&7 of the status register */
	result |= parport_read_status (port) << 8;

	/* Set data bits 5&7 low */
	parport_write_data (port, 0x42);
	/* Read bits 6&7 of the status register */
	result |= parport_read_status (port) << 16;

	/* now check the results */
	switch (result & 0x606060)
	{
	case 0x00004000:
		/* V1 rev 0 Hardware */
		return BB_VER_1 | BB_REV_0;

	case 0x00006000:
		/* V1 rev 1 Hardware */
		return BB_VER_1 | BB_REV_1;

	default:
		return 0;
	}
}

ssize_t byteblaster_read(struct file * file, char * data, size_t length, loff_t * offset)
{
	struct byteblaster_t * bb = file->private_data;
	int remain = length;
	int avail;

	/* Check access to user space */
	if (!access_ok (VERIFY_WRITE, (const void*)data, length))
		return -EFAULT;

	if (down_interruptible (&bb->port_mutex))
		return -EINTR;

	/* 
	 * Copy out data until the buffer empties or the required number of bytes
	 * have been read.  This loop has been carefully designed so that it doesn't
	 * need locking if another thread is putting data into the FIFO at the same
	 * time (ie is incrementing read_fifo_in).
	 */
	while (remain > 0) {
		avail = bb->read_fifo_in - bb->read_fifo_out;
		if (avail == 0)
			break;
		else if (avail < 0)
			avail = READ_FIFO_SIZE - bb->read_fifo_out;

		if (avail > remain)
			avail = remain;

		__copy_to_user (data, bb->read_fifo + bb->read_fifo_out, avail);

		bb->read_fifo_out += avail;
		data += avail;
		if (bb->read_fifo_out == READ_FIFO_SIZE)
			bb->read_fifo_out = 0;

		remain -= avail;
	}

	up (&bb->port_mutex);

	return length - remain;
}

/* Clock some data out of the ByteBlaster.  This function will have undefined
 * results if more than one thread calls write at the same time.  It is safe
 * for one thread to call read while another calls write.
 */

ssize_t byteblaster_write(struct file * file, const char * userdata, size_t length, loff_t* offset)
{
	struct byteblaster_t * const bb = file->private_data;
	struct parport * const port = bb->dev->port;
	size_t remain = length;
	ssize_t err = 0;
	char outbuffer[1024];
	char * outdata;
	char out = 0x42;
	char in = bb->last_in;

	/* 
	 * Write a series of bytes out to the parallel port
	 * 
	 * b0 - don't sample input 
	 * b1 - TMS value
	 * b6 - TDI value
	 * b7 - 1=xor TDI with previous TDO
	 */

	if (down_interruptible (&bb->port_mutex))
		return -EINTR;

	while (remain > 0)
	{
		size_t copy_len;
		char * outend;

		/* get the data from user space */
		copy_len = remain > sizeof(outbuffer) ? sizeof(outbuffer) : remain;
		if (copy_from_user (outbuffer, userdata, copy_len) != 0) {
			if (length == remain)
				err = -EFAULT;
			goto error;
		}

		userdata += copy_len;
		remain   -= copy_len;

		outdata = outbuffer;
		outend  = outbuffer + copy_len;

		for ( ; outdata < outend ; outdata++) {
			out = *outdata;

			if (out & 0x80) {
				/* If 'in' is invalid then make it valid now (we know that it
				 * is always valid at the start of the buffer). */
				if (outdata != outbuffer && (outdata[-1] & 1) != 0)
					in = parport_read_status (port);

				out ^= (~in >> 1) & 0x40;
			}
			out &= 0x42;

			/* set TCK low, TMS/TDI as specified */
			parport_write_data(port, out);

			/* Set TCK high */
			parport_write_data(port, out | 1);

			/* Read RTCK/TDO and store in the read fifo if required */
			if ((*outdata & 1) == 0) {
				unsigned int next = bb->read_fifo_in + 1;
				if (next == READ_FIFO_SIZE)
					next = 0;

				in = parport_read_status (port);

				if (next == bb->read_fifo_out) {
					err = -ENOBUFS;
					goto error;
				}

				bb->read_fifo[bb->read_fifo_in] = in;
				bb->read_fifo_in = next;
			}
		}

		/* Read status just in case it is needed next time */
		in = parport_read_status (port);
	}

error:

	bb->last_in = in;

	/* Set TCK low */
	parport_write_data (port, out);

	up (&bb->port_mutex);

	return (err != 0) ? err : length - remain;
}

int byteblaster_ioctl(struct inode* inode, struct file* file,
		 unsigned int cmd, unsigned long arg)
{
	struct byteblaster_t* bb = file->private_data;
	unsigned int in, out;
	int err;

	switch(cmd){
	case BYTEBLASTER_IOCGETHW:
		out = byteblaster_check_hardware (bb->dev->port);
		return put_user (out, (unsigned int *)arg);

	case BYTEBLASTER_IOCGETVER:
		return put_user (0, (unsigned int *)arg);

	case BYTEBLASTER_IOCDIRECT:
		/* Write bits b0, b1 & b6 to the output 
		 * Return the value read back from the port.
		 */
		err = get_user (out, (unsigned int*)arg);
		if (err < 0)
			return err;

		parport_write_data (bb->dev->port, out & 0x43);
		in = parport_read_status (bb->dev->port);

		/* 
		 * Read back data and put it into *arg. Use __put_user as
		 * access has already been checked by get_user 
		 */
		return __put_user (in,(unsigned int*)arg);

	default:
		return -EINVAL;
	}
}

int byteblaster_open(struct inode * inode, struct file * file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	int err;
	struct byteblaster_t * bb = &byteblaster[minor];
	if (minor >= BYTEBLASTER_MAX || byteblaster[minor].dev == NULL)
		return -ENODEV;
	if (test_and_set_bit(BYTEBLASTER_BUSY_BIT_POS, &bb->flags))
		return -EBUSY;
	/* Claim the parallel port so that we can start using it */
	err = parport_claim (bb->dev);
	if (err < 0) {
		bb->flags &= ~BYTEBLASTER_BUSY;
		return err;
	}

	/* Set bits to idle state and turn on autofeed to enable ByteBlaster */
	parport_write_data    (bb->dev->port, 0x42);
	parport_write_control (bb->dev->port, PARPORT_CONTROL_AUTOFD);
	printk("Byteblaster enable\n");

	/* Check that there is actually a byteblaster connected to this port */
	if (!byteblaster_check_hardware (bb->dev->port)) {
		parport_release (bb->dev);
		bb->flags &= ~BYTEBLASTER_BUSY;
		return -EIO;
	}

	bb->read_fifo = (char *) kmalloc (READ_FIFO_SIZE, GFP_KERNEL);
	if (bb->read_fifo == NULL) {
		parport_release (bb->dev);
		bb->flags &= ~BYTEBLASTER_BUSY;
		return -ENOMEM;
	}

	bb->read_fifo_in  = 0;
	bb->read_fifo_out = 0;
	bb->last_in = 0;

	file->private_data = bb;
	printk("Byteblaster open ok\n");

	return 0;
}

int byteblaster_release(struct inode * inode, struct file * file)
{
	struct byteblaster_t* bb = file->private_data;

	/* Put autofeed low to disable ByteBlaster */
	parport_write_control(bb->dev->port, 0);

	parport_release(bb->dev);

	lock_kernel();
	kfree(bb->read_fifo);
	bb->read_fifo = NULL;
	bb->flags &= ~BYTEBLASTER_BUSY;
	unlock_kernel();

	return 0;
}

struct file_operations byteblaster_fops ={
	owner:		THIS_MODULE,
	read:     byteblaster_read,
	write:    byteblaster_write,
	ioctl:    byteblaster_ioctl,
	open:     byteblaster_open,
	release:  byteblaster_release,
};

static void byteblaster_attach(struct parport *port)
{
	unsigned int i;
	char name[16];

	for (i = 0; i < BYTEBLASTER_MAX; i++)
		if (byteblaster[i].dev == NULL) {
			byteblaster[i].dev = parport_register_device(port, "byteblaster", 
			        NULL, NULL, NULL, 0, (void *) &byteblaster[i]);

			if (byteblaster[i].dev == NULL)
				return;

			sprintf (name, "byteblaster%d", i);
			/*byteblaster[i].devfs_handle = devfs_register (NULL, name,
					DEVFS_FL_DEFAULT, byteblaster_major, i,
					S_IFCHR | S_IRUGO | S_IWUGO,
					&byteblaster_fops, NULL);*/

			printk(KERN_INFO "byteblaster%d: using %s.\n", i, port->name);

			return;
		}
}

static void byteblaster_detach(struct parport *port)
{
	// Ports can't be removed while a device is attached to them.
	// Do we really need this?

	unsigned int i;

	for (i = 0; i < BYTEBLASTER_MAX; i++)
		if (byteblaster[i].dev != NULL && byteblaster[i].dev->port == port) {

			parport_unregister_device(byteblaster[i].dev);
			//devfs_unregister(byteblaster[i].devfs_handle);

			byteblaster[i].dev = NULL;
			//byteblaster[i].devfs_handle = NULL;
		}
}

static struct parport_driver byteblaster_driver = {
	.name = "byteblaster",
	.attach = byteblaster_attach,
	.detach = byteblaster_detach,
};

static int __init byteblaster_init(void)
{
	int i;
	int err;

	for (i = 0 ; i < BYTEBLASTER_MAX ; i++) {
		byteblaster[i].dev = NULL;
		byteblaster[i].flags = 0;
		//byteblaster[i].devfs_handle = NULL;
		init_MUTEX(&byteblaster[i].port_mutex);
		byteblaster[i].read_fifo = NULL;
	}

	err = register_chrdev(byteblaster_major, "byteblaster", &byteblaster_fops);
	if (err < 0)
		return err;

	byteblaster_major = err;

	if (parport_register_driver (&byteblaster_driver)) {
		printk ("byteblaster: unable to register with parport\n");
		return -EIO;
	}

	return 0;
}

static void byteblaster_cleanup_module(void)
{
	int i;

	parport_unregister_driver (&byteblaster_driver);

	printk("Unregister chrdev\n");
	unregister_chrdev(byteblaster_major, "byteblaster");
	printk("Unregister chrdev ok\n");

	for (i = 0 ; i < BYTEBLASTER_MAX ; i++)
		if (byteblaster[i].dev != NULL) {
			parport_unregister_device(byteblaster[i].dev);
			//unregister(byteblaster[i].devfs_handle);
		}
}

module_init(byteblaster_init);
module_exit(byteblaster_cleanup_module);

MODULE_DESCRIPTION ("Driver for the Altera ByteBlaster JTAG cable.");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

